/* Program to facilitate reordering the InstantTrack satellite database */
/* Copyright 1990 Paul Williamson */

/* Compile for Large memory model! */
/* Microsoft C 5.1 command line:
     cl -W3 -AL -Oailt itsort.c		*/
/* Turbo C 2.0 command line:
     tcc -w -ml -O -G itsort.c		*/

/* Microsoft makes a faster .EXE file for this program. */

/***************************************************************\
*                                                               *
* The InstantTrack database begins with a 5-byte version        *
* signature, like "0.85b" for versions from 0.85b to 1.00.      *
* Following that are MAXSAT satellite records.  Each satellite  *
* record begins with the name of the satellite, up to 19        *
* characters plus a null termination.  The remaining portion    *
* of each satellite record contains information about that      *
* satellite.  The details of that information are of no         *
* interest to this program.                                     *
*                                                               *
* PLEASE NOTE:  There is no guarantee that future versions      *
* of InstantTrack will be compatible with this format!          *
*                                                               *
* This program is used in two steps.  First, it is used to      *
* create a list of the satellite names in the current file.     *
* This list is a simple text file, named SATNAMES.LST, with     *
* one satellite name per line.  The user then edits this text   *
* file, placing the satellite names in the desired order.       *
* He must preserve the one-name-per-line format.  Then the      *
* second step is to run the program again.  It will notice      *
* that SATNAMES.LST already exists.  It reads in the entire     *
* database, then writes it back out in the order specified      *
* by SATNAMES.LST.  If a satellite name appears in the list     *
* that is not in the current database, the program creates      *
* a dummy record for it.  The old IT.ORB is renamed to          *
* ITORB.BAK; any old ITORB.BAK is deleted.                      *
*                                                               *
\***************************************************************/

#include <stdio.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <io.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <dos.h>
#include <ctype.h>

#define MAXSAT	200		/* possible version dependency here! */

				/* offset of N0 in satellite record */
#define	N0_OFFSET	(60+2*sizeof(double))
				/* offset of BLAT in satellite record */
#define	BLAT_OFFSET	(60+7*sizeof(double)+3*sizeof(float)+sizeof(long))
#define BLAT_NONE  32767	/*no spacecraft attitude set*/

unsigned int	filesize;
unsigned int	arraysize;
unsigned int	satsize;
char		path[80];

int	main(void);
void	make_list(void);
void	update_database(void);

/************************************************/

int main(void)
{
struct stat	fileinfo;
char		filename[80];
char		*envp;

puts("ITSORT 1.04 - InstantTrack Satellite Database Sorting Utility\n");
puts("Copyright 1990 Paul Williamson, KB5MU\n\n");

		/* figure out what directory the file should be in */

if ((envp = getenv("INSTANTTRACK")) != NULL)
    {
    strncpy(path, envp, 70);
    strcat(path, "\\");
    }
else
    path[0] = '\0';

strcpy(filename, path);			/* look for the file there */
strcat(filename, "IT.ORB");
if (stat(filename, &fileinfo))
    {			      		/* not on the named path */
    path[0] = '\0';			/* try current directory */
    if (stat("IT.ORB",  &fileinfo))
	{
	fputs("Sorry, I can't find IT.ORB anywhere.\n", stderr);
	exit(1);
	}
    }
					/* figure out object sizes */
filesize = (unsigned) fileinfo.st_size;	/* assume < 64K */
arraysize = filesize - 5;
satsize = arraysize / MAXSAT;
arraysize -= 2*satsize;			/* don't include Sun and Moon */

if (access("SATNAMES.LST", 0))		/* does the list already exist? */
    make_list();			/* no, create it. */
else
    update_database();			/* yes, use it to update IT.ORB */

return 0;
}

/**********************************************************\
*                                                          *
* make_list                                                *
* This function reads IT.ORB.  For each satellite record,  *
* it prints the name into the file SATNAMES.LST.           *
*                                                          *
\**********************************************************/

void make_list(void)
{
FILE	*itfile, *lstfile;
char	filename[80];
char	buf[20];
int     i;

strcpy(filename, path);
strcat(filename, "IT.ORB");
if ((itfile = fopen(filename, "rb")) == NULL)
    {
    perror("Can't open IT.ORB");
    exit(1);
    }

if ((lstfile = fopen("SATNAMES.LST", "w")) == NULL)
    {
    perror("Can't create SATNAMES.LST");
    exit(1);
    }

puts("Scanning IT.ORB, and creating SATNAMES.LST ...");

setvbuf(itfile, NULL, _IOFBF, 4096);	/* turn on buffering for speed */
fseek(itfile, 5L+2*satsize, SEEK_CUR);	/* go past version id string and
					/* mandatory Sun and Moon entries */

for (i=0; i < MAXSAT-2; i++)
    {
    fread(buf, 20, 1, itfile);		/* get the name */
    buf[19] = 0;			/* make real sure it's terminated */
    if (*buf)
	{
	fputs(buf, lstfile);		/* write it out */
	fputc('\n', lstfile);		/* one per line */
	}
    fseek(itfile, (long) (satsize-20), SEEK_CUR);/* skip rest of record */
    }

if (ferror(itfile))
    {
    fputs("Error reading IT.ORB", stderr);
    exit(1);
    }

if (ferror(lstfile))
    {
    fputs("Error writing SATNAMES.LST", stderr);
    exit(1);
    }

if (fclose(lstfile))
    {
    perror("Error creating SATNAMES.LST");
    exit(1);
    }

puts("\n"
     "SATNAMES.LST has been created.  Edit SATNAMES.LST with any text\n"
     "editor that creates plain ASCII files.  Rearrange the names in\n"
     "SATNAMES.LST to suit your needs, but make sure there is exactly\n"
     "one name on each line.  Then run this program again to incorporate\n"
     "the changes in the InstantTrack satellite database.");

}

/*********************************************************************\
*                                                                     *
* update_database                                                     *
* This function reads the old IT.ORB (which is renamed to ITORB.BAK)  *
* and the previously created SATNAMES.LST, and writes out a new       *
* IT.ORB file.  It copies the version string and the Sun and Moon     *
* records from the old IT.ORB.  Then for each name in SATNAMES.LST    *
* it finds the old record in memory and writes it to the new IT.ORB.  *
* For names in SATNAMES.LST that didn't exist in the old database,    *
* it creates a dummy record.  It pads the end of the new IT.ORB with  *
* blank records.                                                      *
*                                                                     *
\*********************************************************************/

void update_database(void)
{
FILE	*lstfile, *oldfile, *newfile;
char	oldfn[80], newfn[80];
char	*array, *p, *dummysat;
char	satname[20], dummy[20];
char	copyflag[MAXSAT];
int	newslot, oldslot;
int	copycount = 0;
int	dummycount = 0;
int	found;
int	drive;
long	freespace;
#ifdef __TURBOC__
struct dfree diskinfo;
#else
struct diskfree_t diskinfo;
#endif

if (((dummysat = (char *)calloc(1, satsize)) == NULL) ||
           ((array = (char *)malloc(arraysize)) == NULL))
    {
    fputs("Couldn't get enough memory to hold the database.\n", stderr);
    exit(1);
    }
else
    {
    *(double *)(dummysat+N0_OFFSET) = 0.01;/* IT 1.00 can't handle N0=0 */
    *(int *)(dummysat+BLAT_OFFSET) = BLAT_NONE;	/* no attitude set */
    }

if ((lstfile = fopen("SATNAMES.LST", "r")) == NULL)
    {
    perror("Can't read SATNAMES.LST");
    exit(1);
    }

strcpy(oldfn, path);
strcat(oldfn, "ITORB.BAK");
strcpy(newfn, path);
strcat(newfn, "IT.ORB");

if (access(oldfn, 0) == 0)		/* check for old ITORB.BAK */
    {
    if (unlink(oldfn))			/* delete it */
	{
	perror("Couldn't get rid of old ITORB.BAK file");
	exit(1);
	}
    }
else
    {					/* make sure there's free space */
    if (path[0] && (path[1] == ':'))	/* on the relevant drive */
	drive = toupper(path[0]) - 'A' + 1;
    else
	drive = 0;
#ifdef __TURBOC__
    getdfree(drive, &diskinfo);
    if (diskinfo.df_sclus == 0xFFFF)
#else
    if (_dos_getdiskfree(drive, &diskinfo))
#endif
	{
	fputs("Bad drive designator in INSTANTTRACK path.", stderr);
	exit(1);
	}

#ifdef __TURBOC__
    freespace = (long) diskinfo.df_avail *
		(long) diskinfo.df_sclus *
		(long) diskinfo.df_bsec;
#else
    freespace = (long) diskinfo.avail_clusters * 
		(long) diskinfo.sectors_per_cluster *
		(long) diskinfo.bytes_per_sector;
#endif

    if (freespace < filesize)
	{
	fputs("Not enough free disk space to create backup file.  Aborting.", stderr);
	exit(1);
	}
    }

if (rename(newfn, oldfn))		/* move IT.ORB to ITORB.BAK */
    {
    perror("Couldn't rename IT.ORB to ITORB.BAK");
    exit(1);
    }    

if ((oldfile = fopen(oldfn, "rb")) == NULL)
    {
    perror("Renamed IT.ORB to ITORB.BAK, but then couldn't read it");
    exit(1);
    }

if ((newfile = fopen(newfn, "wb")) == NULL)
    {
    perror("Renamed IT.ORB to ITORB.BAK, but couldn't make a new IT.ORB");
    exit(1);
    }

puts("Updating the InstantTrack satellite database ...\n");

setvbuf(oldfile, NULL, _IOFBF, 4096);		/* buffering for speed */
setvbuf(newfile, NULL, _IOFBF, 4096);

fread(array, 5+2*satsize, 1, oldfile);		/* copy version, Sun, Moon */
fwrite(array, 5+2*satsize, 1, newfile);

fread(array, arraysize, 1, oldfile);		/* read in everything else */

for (oldslot=0; oldslot < MAXSAT-2; oldslot++)
    copyflag[oldslot] = 0;

for (newslot=0; newslot < MAXSAT-2; newslot++)
    {
    if (fgets(satname, 20, lstfile) == NULL)
	break;

    if ((p=strchr(satname, '\n')) != NULL)	/* trim newline */
	while (isspace(*p) && (p >= satname))	/* trim trailing whitespace */
	    *p-- = '\0';
    else
	{
	fputs("Warning: long line in SATNAMES.LST.  Truncated to:\n", stderr);
	fputs(satname, stderr);
	fputc('\n', stderr);
	while (fgets(dummy, 20, lstfile) && ! strchr(dummy, '\n'))
	    ;		/* discard characters up to next newline */
	}

    found = 0;
    for ((p=array), oldslot=0; oldslot < MAXSAT-2; oldslot++, p+=satsize)
	if (strnicmp(satname, p, 20) == 0)
	    {
	    strncpy(p, satname, 20);/* incorporate any capitalization changes */
	    if (copyflag[oldslot] != 0)		/* Duplicate? */
		p[20] = '\0';			/* scrap Object # */
	    fwrite(p, satsize, 1, newfile);	/* copy modified record */
	    copycount++;			/* count it */
	    copyflag[oldslot] += 1;		/* flag it copied */
	    found++;				/* note that we found it */
	    break;
	    }

    if (!found)			/* if not found, create a dummy satellite */
	{
	strcpy(dummysat, satname);
	fwrite(dummysat, satsize, 1, newfile);
	dummycount++;
	}
    } /* for newslot */

dummysat[0] = '\0';		/* make dummy null again */
for (/* old newslot */; newslot < MAXSAT-2; newslot++)
    fwrite(dummysat, satsize, 1, newfile);	/* pad end of IT.ORB */

if (ferror(newfile))
    {
    perror("Error writing new IT.ORB file");
    exit(1);
    }

if (fclose(newfile))
    {
    perror("Error creating new IT.ORB file");
    exit(1);
    }

printf("Copied %d satellites from old IT.ORB.\n", copycount);
if (dummycount)
    printf("Created %d dummy satellites with new names.\n", dummycount);

for ((p=array), oldslot=0; oldslot < MAXSAT-2; oldslot++, p+=satsize)
    if (*p)
        if (copyflag[oldslot] == 0)
	    printf("Deleted satellite %s.\n", p);
        else if (copyflag[oldslot] == 2)
            printf("Duplicated satellite %s.\n", p);
        else if (copyflag[oldslot] > 2)
	    printf("Duplicated satellite %s %d times.\n", p,
			copyflag[oldslot] - 1);

puts("\nIT.ORB has been updated.");

if (unlink("SATNAMES.LST"))
    perror("Couldn't delete SATNAMES.LST");
else
    puts("SATNAMES.LST file has been deleted.");
}

